/* -*- buffer-read-only: t -*- vi: set ro:
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
*/
#define MF_SOURCE_NAME  MF_SOURCE_BI_DB
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include "mailfromd.h"
#include "prog.h"
#include "builtin.h"

#line 1 "bi_db.m4"
/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008, 2009 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define DEFAULT_DB_MODE 0640
#include <fnmatch.h>

struct db_prop {                /* Database properties */
	char *pat;              /* Database name pattern */
	mode_t mode;            /* File mode */
	int null;               /* Null byte */
};

static mu_list_t db_prop_list;

static int
db_prop_compare(const void *item, const void *data)
{
	const struct db_prop *prop = item;
	const char *name = data;
	return strcmp(prop->pat, name);
}

static int
strtomode(char *str, mode_t *pmode)
{
	mode_t mode = 0;
	struct { char c; unsigned mask; } modetab[] = {
		{ 'r', S_IRUSR },
		{ 'w', S_IWUSR },
		{ 'x', S_IXUSR },
		{ 'r', S_IRGRP },
		{ 'w', S_IWGRP },
		{ 'x', S_IXGRP },
		{ 'r', S_IROTH },
		{ 'w', S_IWOTH },
		{ 'x', S_IXOTH },
		{ 0 }
	};
	int i;

	for (i = 0; modetab[i].c; i++) {
		if (!str[i])
			return i + 1;
		if (str[i] == modetab[i].c)
			mode |= modetab[i].mask;
		else if (str[i] != '-')
			return i + 1;
	}
	if (str[i])
		return i + 1;
	*pmode = mode;
	return 0;
}

/* #pragma dbprop <name> [null] [mode] */
#line 69 "bi_db.m4"

#line 69
static void _pragma_dbprop (int argc, char **argv, const char *text)
#line 69

{
	int null = 0;
	mode_t mode = DEFAULT_DB_MODE;
	struct db_prop *prop;
	int rc;
	char *pat;

	--argc;
	pat = *++argv;
	
	while (--argc) {
		char *p = *++argv;

		if (strcmp (p, "null") == 0)
			null = 1;
		else if (c_isdigit(*p)) {
			unsigned long n = strtoul(p, &p, 8);
			if (*p || (mode = n) != n) {
				parse_error(_("bad numeric file mode"));
				return;
			}
		} else if (rc = strtomode(p, &mode)) {
			parse_error(_("bad symbolic file mode (near %s)"),
				    p + rc - 1);
			return;
		}
	}

	if (!db_prop_list) {
		rc = mu_list_create(&db_prop_list);
		if (rc) {
			parse_error(_("Cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
		mu_list_set_comparator(db_prop_list, db_prop_compare);
	}

	if (mu_list_locate(db_prop_list, pat, (void**) &prop)) {
		prop = xmalloc(sizeof(*prop));
		prop->pat = xstrdup(pat);
		rc = mu_list_append(db_prop_list, prop);
		if (rc) {
			parse_error(_("Cannot create list: %s"),
				    mu_strerror(rc));
			return;
		}
	}
	prop->mode = mode;
	prop->null = null;
}

const struct db_prop *
db_prop_lookup(const char *name)
{
	mu_iterator_t itr;
	const struct db_prop *found = NULL;
	
	if (db_prop_list && mu_list_get_iterator(db_prop_list, &itr)) {
		for (mu_iterator_first(itr);
		     !mu_iterator_is_done(itr);
		     mu_iterator_next(itr)) {
			const struct db_prop *p;
			mu_iterator_current(itr, (void**)&p);
			if (strcmp(p->pat, name) == 0) {
				found = p;
				break;
			} else if (fnmatch(p->pat, name, 0) == 0)
				found = p;
		}
		mu_iterator_destroy(&itr);
	}
	return found;	
}


#define LOOKUP_NULL_BYTE 0x1
#define LOOKUP_TEST_ONLY 0x2

static int
dbmap_lookup(eval_environ_t env, char *dbname, const char *keystr,
	     const char *defval, int flags)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;

	if (!defval)
		defval = "";
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 165

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = (void*) keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (flags & LOOKUP_NULL_BYTE)
		MU_DATUM_SIZE(key)++;
	rc = mu_dbm_fetch(&db, key, &contents) == 0;
	debug2(10, "Looking up %s: %s", keystr, rc ? "true" : "false");
	if (flags & LOOKUP_TEST_ONLY) 
		push(env, (STKVAL)rc);
	else {
		if (!rc) {
			if (defval)
				pushs(env, (STKVAL)defval);
			else
				push(env, 0);
		} else if (((char*)MU_DATUM_PTR(contents))[MU_DATUM_SIZE(contents)-1]) {
			size_t off;
			size_t len = MU_DATUM_SIZE(contents);
			char *s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
			memcpy(s, MU_DATUM_PTR(contents), len);
			s[len] = 0;
			push(env, (STKVAL) off);
		} else
			pushs(env, MU_DATUM_PTR(contents));
	}
	mu_dbm_datum_free(&contents);
	mu_dbm_close(&db);
	return rc;
}

void
#line 197
bi_dbmap(eval_environ_t env)
#line 197

#line 197 "bi_db.m4"
{
#line 197
	
#line 197

#line 197
long __bi_argcnt;
#line 197
char * dbname;
#line 197
        char * key;
#line 197
        long  null;
#line 197
        
#line 197
get_string_arg(env, 1, &dbname);
#line 197
        get_string_arg(env, 2, &key);
#line 197
        get_numeric_arg(env, 3, &null);
#line 197
        
#line 197
get_numeric_arg(env, 0, &__bi_argcnt);
#line 197
        adjust_stack(env, __bi_argcnt + 1);
#line 197

#line 197
	if (prog_trace_option & BUILTIN_MASK_db)
#line 197
		prog_trace(env, "dbmap %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? null : 0));;
#line 197

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, NULL,
		     LOOKUP_TEST_ONLY | 
		         (((__bi_argcnt > 2) ? null : prop && prop->null)
			   ? LOOKUP_NULL_BYTE : 0));
}
}

void
#line 207
bi_dbget(eval_environ_t env)
#line 207

#line 207 "bi_db.m4"
{
#line 207
	
#line 207

#line 207
long __bi_argcnt;
#line 207
char * dbname;
#line 207
        char * key;
#line 207
        char * defval;
#line 207
        long  null;
#line 207
        
#line 207
get_string_arg(env, 1, &dbname);
#line 207
        get_string_arg(env, 2, &key);
#line 207
        get_string_arg(env, 3, &defval);
#line 207
        get_numeric_arg(env, 4, &null);
#line 207
        
#line 207
get_numeric_arg(env, 0, &__bi_argcnt);
#line 207
        adjust_stack(env, __bi_argcnt + 1);
#line 207

#line 207
	if (prog_trace_option & BUILTIN_MASK_db)
#line 207
		prog_trace(env, "dbget %s %s %s %lu",dbname, key, ((__bi_argcnt > 2) ? defval : ""), ((__bi_argcnt > 3) ? null : 0));;

{
	const struct db_prop *prop = db_prop_lookup(dbname);
	dbmap_lookup(env, dbname, key, 
                     ((__bi_argcnt > 2) ? defval : 0), 
		     ((__bi_argcnt > 3) ? null : prop && prop->null)
		      ? LOOKUP_NULL_BYTE : 0);
}
}

void
#line 218
bi_dbput(eval_environ_t env)
#line 218

#line 218 "bi_db.m4"
{
#line 218
	
#line 218

#line 218
long __bi_argcnt;
#line 218
char * dbname;
#line 218
        char * keystr;
#line 218
        char * value;
#line 218
        long  null;
#line 218
        long  mode;
#line 218
        
#line 218
get_string_arg(env, 1, &dbname);
#line 218
        get_string_arg(env, 2, &keystr);
#line 218
        get_string_arg(env, 3, &value);
#line 218
        get_numeric_arg(env, 4, &null);
#line 218
        get_numeric_arg(env, 5, &mode);
#line 218
        
#line 218
get_numeric_arg(env, 0, &__bi_argcnt);
#line 218
        adjust_stack(env, __bi_argcnt + 1);
#line 218

#line 218
	if (prog_trace_option & BUILTIN_MASK_db)
#line 218
		prog_trace(env, "dbput %s %s %s %lu %lu",dbname, keystr, value, ((__bi_argcnt > 3) ? null : 0), ((__bi_argcnt > 4) ? mode : 0));;

{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR,
			((__bi_argcnt > 4) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
			env_throw_bi(env, mfe_dbfailure, "dbput", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 234
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (((__bi_argcnt > 3) ? null : prop && prop->null))
		MU_DATUM_SIZE(key)++;

	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(contents) = value;
	MU_DATUM_SIZE(contents) = strlen(value) + 1;
	
	rc = mu_dbm_insert(&db, key, contents, 1);
	mu_dbm_close(&db);
		if (!(rc == 0))
#line 246
			env_throw_bi(env, mfe_dbfailure, "dbput", _("Failed to insert data to %s: %s %s: %s"),dbname,keystr,value,mu_dbm_strerror())
#line 252
;
}
}

void
#line 256
bi_dbdel(eval_environ_t env)
#line 256

#line 256 "bi_db.m4"
{
#line 256
	
#line 256

#line 256
long __bi_argcnt;
#line 256
char * dbname;
#line 256
        char * keystr;
#line 256
        long  null;
#line 256
        long  mode;
#line 256
        
#line 256
get_string_arg(env, 1, &dbname);
#line 256
        get_string_arg(env, 2, &keystr);
#line 256
        get_numeric_arg(env, 3, &null);
#line 256
        get_numeric_arg(env, 4, &mode);
#line 256
        
#line 256
get_numeric_arg(env, 0, &__bi_argcnt);
#line 256
        adjust_stack(env, __bi_argcnt + 1);
#line 256

#line 256
	if (prog_trace_option & BUILTIN_MASK_db)
#line 256
		prog_trace(env, "dbdel %s %s %lu %lu",dbname, keystr, ((__bi_argcnt > 2) ? null : 0), ((__bi_argcnt > 3) ? mode : 0));;

{
	DBM_FILE db;
	DBM_DATUM key;
	int rc;
	const struct db_prop *prop = db_prop_lookup(dbname);
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_RDWR,
			((__bi_argcnt > 3) ? mode : (prop ? prop->mode : DEFAULT_DB_MODE)),
			NULL))
			env_throw_bi(env, mfe_dbfailure, "dbdel", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 271
	memset(&key, 0, sizeof key);
	MU_DATUM_PTR(key) = keystr;
	MU_DATUM_SIZE(key) = strlen(keystr);
	if (((__bi_argcnt > 2) ? null : prop && prop->null))
		MU_DATUM_SIZE(key)++;
	rc = mu_dbm_delete(&db, key);
	mu_dbm_close(&db);
		if (!(rc == 0))
#line 278
			env_throw_bi(env, mfe_dbfailure, "dbdel", _("Failed to delete data `%s' from `%s': %s"),keystr,dbname,mu_dbm_strerror())
#line 283
;
}
}


#define NUMDB 128

struct db_tab {
	int used;
	DBM_FILE db;
	DBM_DATUM key;
};

static void *
alloc_db_tab()
{
	return xcalloc(NUMDB, sizeof(struct db_tab));
}

static void
close_db_tab(struct db_tab *dbt)
{
	if (dbt->used) {
		mu_dbm_datum_free(&dbt->key);
		mu_dbm_close(&dbt->db);
		dbt->used = 0;
	}
}

static void
destroy_db_tab(void *data)
{
	int i;
	struct db_tab *db = data;
	for (i = 0; i < NUMDB; i++) 
		close_db_tab(db + i);
	free(db);
}


#line 322

#line 322
static int DBTAB_id;
#line 322 "bi_db.m4"
;

static int
new_db_tab(struct db_tab *dbt)
{
	int i;
	for (i = 0; i < NUMDB; i++)
		if (!dbt[i].used) {
			dbt[i].used = 1;
			return i;
		}
	return -1;
}
			


void
#line 338
bi_dbfirst(eval_environ_t env)
#line 338

#line 338 "bi_db.m4"
{
#line 338
	
#line 338

#line 338

#line 338
char * dbname;
#line 338
        
#line 338
get_string_arg(env, 0, &dbname);
#line 338
        
#line 338

#line 338
        adjust_stack(env, 1);
#line 338

#line 338
	if (prog_trace_option & BUILTIN_MASK_db)
#line 338
		prog_trace(env, "dbfirst %s",dbname);;
#line 338

{
	int rc;
	int n;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id);
	DBM_FILE db;
	DBM_DATUM key;
	
	if (mu_dbm_open(dbname, &db, MU_STREAM_READ, 0, NULL))
			env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mu_dbm_open(%s) failed: %s"),dbname,mu_dbm_strerror());
#line 351
	rc = mu_dbm_firstkey(&db, &key);
		if (!(rc == 0))
#line 352
			env_throw_bi(env, mfe_dbfailure, "dbfirst", _("mu_dbm_firstkey failed: %s"),mu_dbm_strerror())
#line 354
;
	n = new_db_tab(dbt);
		if (!(n >= 0))
#line 356
			env_throw_bi(env, mfe_failure, "dbfirst", _("No more database entries available"))
#line 358
;
	dbt += n;
	dbt->db = db;
	dbt->key = key;
	
#line 362
do {
#line 362
  push(env, (STKVAL) (n));
#line 362
  return;
#line 362
} while (0);
}
}

void
#line 366
bi_dbnext(eval_environ_t env)
#line 366

#line 366 "bi_db.m4"
{
#line 366
	
#line 366

#line 366

#line 366
long  dn;
#line 366
        
#line 366
get_numeric_arg(env, 0, &dn);
#line 366
        
#line 366

#line 366
        adjust_stack(env, 1);
#line 366

#line 366
	if (prog_trace_option & BUILTIN_MASK_db)
#line 366
		prog_trace(env, "dbnext %lu",dn);;
#line 366

{
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	DBM_DATUM nextkey;
	int rc;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 372
			env_throw_bi(env, mfe_range, "dbnext", _("Invalid database descriptor"))
#line 374
;

	rc = mu_dbm_nextkey(&dbt->db, dbt->key, &nextkey);
	if (rc) {
		close_db_tab(dbt);
		
#line 379
do {
#line 379
  push(env, (STKVAL) (0));
#line 379
  return;
#line 379
} while (0);
	}
	mu_dbm_datum_free(&dbt->key);
	dbt->key = nextkey;
	
#line 383
do {
#line 383
  push(env, (STKVAL) (1));
#line 383
  return;
#line 383
} while (0);
}
}

void
#line 387
bi_dbkey(eval_environ_t env)
#line 387

#line 387 "bi_db.m4"
{
#line 387
	
#line 387

#line 387

#line 387
long  dn;
#line 387
        
#line 387
get_numeric_arg(env, 0, &dn);
#line 387
        
#line 387

#line 387
        adjust_stack(env, 1);
#line 387

#line 387
	if (prog_trace_option & BUILTIN_MASK_db)
#line 387
		prog_trace(env, "dbkey %lu",dn);;
#line 387

{
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 393
			env_throw_bi(env, mfe_range, "dbkey", _("Invalid database descriptor"))
#line 395
;
	
	len = MU_DATUM_SIZE(dbt->key);
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, MU_DATUM_PTR(dbt->key), len);
	s[len] = 0;
	
#line 401
do {
#line 401
  push(env, (STKVAL) (off));
#line 401
  return;
#line 401
} while (0);
}
}

void
#line 405
bi_dbvalue(eval_environ_t env)
#line 405

#line 405 "bi_db.m4"
{
#line 405
	
#line 405

#line 405

#line 405
long  dn;
#line 405
        
#line 405
get_numeric_arg(env, 0, &dn);
#line 405
        
#line 405

#line 405
        adjust_stack(env, 1);
#line 405

#line 405
	if (prog_trace_option & BUILTIN_MASK_db)
#line 405
		prog_trace(env, "dbvalue %lu",dn);;
#line 405

{
	int rc;
	size_t off, len;
	char *s;
	struct db_tab *dbt = env_get_builtin_priv(env,DBTAB_id) + dn;
	DBM_DATUM contents;
	
		if (!(dn >= 0 && dn < NUMDB && dbt->used))
#line 413
			env_throw_bi(env, mfe_range, "dbvalue", _("Invalid database descriptor"))
#line 415
;

	memset(&contents, 0, sizeof contents);
	rc = mu_dbm_fetch(&dbt->db, dbt->key, &contents);
		if (!(rc == 0))
#line 419
			env_throw_bi(env, mfe_dbfailure, "dbvalue", _("Key not found: %s"),mu_dbm_strerror())
#line 422
;
	
	len = MU_DATUM_SIZE(contents);
	s = (char*) env_data_ref(env, (off = heap_reserve(env, len + 1)));
	memcpy(s, MU_DATUM_PTR(contents), len);
	s[len] = 0;
	mu_dbm_datum_free(&contents);
	
#line 429
do {
#line 429
  push(env, (STKVAL) (off));
#line 429
  return;
#line 429
} while (0);
}
}


enum greylist_semantics
{
	greylist_traditional,
	greylist_ct
};

static enum greylist_semantics greylist_semantics = greylist_traditional;

/* #pragma greylist {traditional|gray|ct|con-tassios}*/
#line 443 "bi_db.m4"

#line 443
static void _pragma_greylist (int argc, char **argv, const char *text)
#line 443

{
	if (strcmp(argv[1], "traditional") == 0
	    || strcmp(argv[1], "gray") == 0)
		greylist_semantics = greylist_traditional;
	else if (strcmp(argv[1], "ct") == 0
		 || strcmp(argv[1], "con-tassios") == 0)
		greylist_semantics = greylist_ct;
	else
		/* TRANSLATORS: Do not translate keywords:
		   traditional, gray, ct, con-tassios */
		parse_error(_("unknown semantics; allowed values are: "
			      "traditional (or gray) and "
			      "ct (or con-tassios)"));
}

static void
greylist_print_item(const char *key, size_t size, const void *content)
{
	time_t timestamp = *(time_t*) content;
	size--; /* Size includes the trailing nul */
	printf("%*.*s ", size, size, key);
	format_time_str(stdout, timestamp);
	putchar('\n');
}

static int
greylist_expire_item(const void *content)
{
	time_t timestamp = *(time_t*) content;
	return greylist_format->expire_interval &&
		time(NULL) - timestamp > greylist_format->expire_interval;
}


static struct db_format greylist_format_struct = {
	"greylist",
	DEFAULT_GREYLIST_DATABASE,
	1,
	DEFAULT_EXPIRE_INTERVAL,
	greylist_print_item,
	greylist_expire_item
};

struct db_format *greylist_format = &greylist_format_struct;

static size_t greylist_seconds_left_loc
#line 489 "bi_db.m4"
;

/* The traditional (aka gray's) greylist implementation: the greylist
   database keeps the time the greylisting was activated.
 */
static int
do_greylist_traditional(eval_environ_t env, char *email, long interval)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
        	if (!(rc == 0))
#line 506
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_dbm_strerror())
;
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp, diff;

			if (!(MU_DATUM_SIZE(contents) == sizeof timestamp))
#line 518
			env_throw_bi(env, mfe_dbfailure, NULL, _("Greylist database %s has wrong data size"),greylist_format->dbname)
#line 521
;
		
		timestamp = *(time_t*) MU_DATUM_PTR(contents);
		diff = now - timestamp;

		__DBG(20) {
			char timebuf[32];
			debug_log("%s entered greylist database on %s, "
				  "%ld seconds ago",
				  email,
				  mailfromd_timestr(timestamp, timebuf,
						    sizeof timebuf),
				  (long) diff);
		}

		if (diff < interval) {
			diff = interval - diff;

			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (diff);
			
			debug2(20, "%s still greylisted (for %lu sec.)",
			       email,
			       (unsigned long) diff);
			rc = 1;
		} else if (diff > greylist_format->expire_interval) {
			debug1(20, "greylist record for %s expired",
			       email);
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
			if (!readonly) {
				memcpy(MU_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mu_dbm_insert(&db, key, contents, 1))
					mu_error(_("Cannot insert datum `%-.*s' in "
						   "greylist database %s: %s"),
			                         MU_DATUM_SIZE(key),
						 (char*)MU_DATUM_PTR(key),
						 greylist_format->dbname,
	                                         mu_dbm_strerror());
			} else
				debug(20, "database opened in readonly mode: "
				      "not updating");
			rc = 1;
		} else {
			debug1(20, "%s finished greylisting period",
			       email);
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		debug1(20, "greylisting %s", email);
		*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
		MU_DATUM_PTR(contents) = (void*)&now;
		MU_DATUM_SIZE(contents) = sizeof now;
		if (mu_dbm_insert(&db, key, contents, 1))
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 MU_DATUM_SIZE(key), (char*)MU_DATUM_PTR(key),
				 greylist_format->dbname,
	                         mu_dbm_strerror());
		rc = 1;
	} else
		rc = 0;

	mu_dbm_close(&db);

	return rc;
}

/* Implementation of the is_greylisted predicate has no sense for 
   traditional greylist databases, because greylisting interval is
   not known beforehand.
   FIXME: keep the reference below up to date.
 */
static int
is_greylisted_traditional(eval_environ_t env, char *email)
{
		env_throw_bi(env, mfe_failure, NULL, _("is_greylisted is not implemented for traditional greylist databases; "
		   "see documentation, chapter %s %s for more info"),"4.12.1.23","Greylisting functions");
#line 601
	return 0;
}

/* New greylist implementation (by Con Tassios): the database keeps
   the time the greylisting period is set to expire (`interval' seconds
   from now)
*/
static int
do_greylist_ct(eval_environ_t env, char *email, long interval)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
        	if (!(rc == 0))
#line 620
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_dbm_strerror())
;
	
	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email)+1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp;

			if (!(MU_DATUM_SIZE(contents) == sizeof timestamp))
#line 632
			env_throw_bi(env, mfe_dbfailure, NULL, _("Greylist database %s has wrong data size"),greylist_format->dbname)
#line 635
;
		
		timestamp = *(time_t*) MU_DATUM_PTR(contents);

		if (now < timestamp) {
			time_t diff = timestamp - now;
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (diff);
			
			debug2(20, "%s still greylisted (for %lu sec.)",
			       email,
			       (unsigned long) diff);
			rc = 1;
		} else if (now - timestamp >
			   greylist_format->expire_interval) {
			debug1(20, "greylist record for %s expired", email);
			*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
			if (!readonly) {
				now += interval;
				memcpy(MU_DATUM_PTR(contents),
				       &now, sizeof now);
				if (mu_dbm_insert(&db, key, contents, 1))
					mu_error(_("Cannot insert datum "
						   "`%-.*s' in greylist "
						   "database %s: %s"),
			                         MU_DATUM_SIZE(key),
						 (char*)MU_DATUM_PTR(key),
						 greylist_format->dbname,
	                                         mu_dbm_strerror());
			} else
				debug(20, "database opened in readonly mode: "
				      "not updating");
			rc = 1;
		} else {
			debug1(20, "%s finished greylisting period", email);
			rc = 0;
		}
		mu_dbm_datum_free(&contents);
	} else if (!readonly) {
		debug1(20, "greylisting %s", email);
		*env_data_ref(env, greylist_seconds_left_loc) = (STKVAL) (interval);
		now += interval;
		MU_DATUM_PTR(contents) = (void*)&now;
		MU_DATUM_SIZE(contents) = sizeof now;
		if (mu_dbm_insert(&db, key, contents, 1))
			mu_error(_("Cannot insert datum `%-.*s' in greylist "
				   "database %s: %s"),
		                 MU_DATUM_SIZE(key), (char*)MU_DATUM_PTR(key),
				 greylist_format->dbname,
	                         mu_dbm_strerror());
		rc = 1;
	} else
		rc = 0;
	
	mu_dbm_close(&db);
	
	return rc;
}

/* The `is_greylisted' predicate for new databases */
static int
is_greylisted_ct(eval_environ_t env, char *email)
{
	int rc;
	DBM_FILE db;
	DBM_DATUM key;
	DBM_DATUM contents;
	int readonly;
	time_t now;

	rc = mu_dbm_open(greylist_format->dbname, &db, MU_STREAM_RDWR, 0600,
			 &readonly);
		if (!(rc == 0))
#line 706
			env_throw_bi(env, mfe_dbfailure, NULL, _("mu_dbm_open(%s) failed: %s"),greylist_format->dbname,mu_dbm_strerror())
;

	memset(&key, 0, sizeof key);
	memset(&contents, 0, sizeof contents);
	MU_DATUM_PTR(key) = email;
	MU_DATUM_SIZE(key) = strlen(email) + 1;

	time(&now);
	if (mu_dbm_fetch(&db, key, &contents) == 0) {
		time_t timestamp;

			if (!(MU_DATUM_SIZE(contents) == sizeof timestamp))
#line 718
			env_throw_bi(env, mfe_dbfailure, NULL, _("Greylist database %s has wrong data size"),greylist_format->dbname)
#line 721
;
                
		timestamp = *(time_t*) MU_DATUM_PTR(contents);
                
		rc = timestamp > now;
                
		mu_dbm_datum_free(&contents);
	} else  
		rc = 0;

	mu_dbm_close(&db);

	return rc;
}

struct greylist_class {
	int (*gl_fun)(eval_environ_t, char *, long);
	int (*gl_pred)(eval_environ_t, char *);
};

struct greylist_class greylist_class[] = {
	{ do_greylist_traditional, is_greylisted_traditional },
	{ do_greylist_ct, is_greylisted_ct }
};

/* greylist(key, interval)

   Returns true if the key is greylisted, false if it's OK to
   deliver mail.
 */
void
#line 751
bi_greylist(eval_environ_t env)
#line 751

#line 751 "bi_db.m4"
{
#line 751
	
#line 751

#line 751

#line 751
char * email;
#line 751
        long  interval;
#line 751
        
#line 751
get_string_arg(env, 0, &email);
#line 751
        get_numeric_arg(env, 1, &interval);
#line 751
        
#line 751

#line 751
        adjust_stack(env, 2);
#line 751

#line 751
	if (prog_trace_option & BUILTIN_MASK_db)
#line 751
		prog_trace(env, "greylist %s %lu",email, interval);;
#line 751

{
	
#line 753
do {
#line 753
  push(env, (STKVAL) (greylist_class[greylist_semantics].gl_fun(env, email,
#line 753
							    interval)));
#line 753
  return;
#line 753
} while (0);
#line 755
}
}

/* is_greylisted(key)
   
   Returns true if the key is greylisted, otherwise false
   
*/
void
#line 763
bi_is_greylisted(eval_environ_t env)
#line 763

#line 763 "bi_db.m4"
{
#line 763
	
#line 763

#line 763

#line 763
char * email;
#line 763
        
#line 763
get_string_arg(env, 0, &email);
#line 763
        
#line 763

#line 763
        adjust_stack(env, 1);
#line 763

#line 763
	if (prog_trace_option & BUILTIN_MASK_db)
#line 763
		prog_trace(env, "is_greylisted %s",email);;
#line 763

{
	
#line 765
do {
#line 765
  push(env, (STKVAL) (greylist_class[greylist_semantics].gl_pred(env, email)));
#line 765
  return;
#line 765
} while (0);
}
}

void
#line 769
bi_db_name(eval_environ_t env)
#line 769

#line 769 "bi_db.m4"
{
#line 769
	
#line 769

#line 769

#line 769
char * fmtid;
#line 769
        
#line 769
get_string_arg(env, 0, &fmtid);
#line 769
        
#line 769

#line 769
        adjust_stack(env, 1);
#line 769

#line 769
	if (prog_trace_option & BUILTIN_MASK_db)
#line 769
		prog_trace(env, "db_name %s",fmtid);;
#line 769

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 772
			env_throw_bi(env, mfe_not_found, "db_name", _("No such db format: %s"),fmtid)
#line 774
;
	do {pushs(env, fmt->dbname);
#line 775
return;
#line 775
} while (0);
}
}

void
#line 779
bi_db_get_active(eval_environ_t env)
#line 779

#line 779 "bi_db.m4"
{
#line 779
	
#line 779

#line 779

#line 779
char * fmtid;
#line 779
        
#line 779
get_string_arg(env, 0, &fmtid);
#line 779
        
#line 779

#line 779
        adjust_stack(env, 1);
#line 779

#line 779
	if (prog_trace_option & BUILTIN_MASK_db)
#line 779
		prog_trace(env, "db_get_active %s",fmtid);;
#line 779

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 782
			env_throw_bi(env, mfe_not_found, "db_get_active", _("No such db format: %s"),fmtid)
#line 784
;
	
#line 785
do {
#line 785
  push(env, (STKVAL) (fmt->enabled));
#line 785
  return;
#line 785
} while (0);
}
}

void
#line 789
bi_db_set_active(eval_environ_t env)
#line 789

#line 789 "bi_db.m4"
{
#line 789
	
#line 789

#line 789

#line 789
char * fmtid;
#line 789
        long  active;
#line 789
        
#line 789
get_string_arg(env, 0, &fmtid);
#line 789
        get_numeric_arg(env, 1, &active);
#line 789
        
#line 789

#line 789
        adjust_stack(env, 2);
#line 789

#line 789
	if (prog_trace_option & BUILTIN_MASK_db)
#line 789
		prog_trace(env, "db_set_active %s %lu",fmtid, active);;
#line 789

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 792
			env_throw_bi(env, mfe_not_found, "db_set_active", _("No such db format: %s"),fmtid)
#line 794
;
	fmt->enabled = active;
}
}

void
#line 799
bi_db_expire_interval(eval_environ_t env)
#line 799

#line 799 "bi_db.m4"
{
#line 799
	
#line 799

#line 799

#line 799
char * fmtid;
#line 799
        
#line 799
get_string_arg(env, 0, &fmtid);
#line 799
        
#line 799

#line 799
        adjust_stack(env, 1);
#line 799

#line 799
	if (prog_trace_option & BUILTIN_MASK_db)
#line 799
		prog_trace(env, "db_expire_interval %s",fmtid);;
#line 799

{
	struct db_format *fmt = db_format_lookup(fmtid);
		if (!(fmt != NULL))
#line 802
			env_throw_bi(env, mfe_not_found, "db_expire_interval", _("No such db format: %s"),fmtid)
#line 804
;
	
#line 805
do {
#line 805
  push(env, (STKVAL) (fmt->expire_interval));
#line 805
  return;
#line 805
} while (0);
}
}


#line 809
void
#line 809
db_init_builtin()
#line 809
{
#line 809
	#line 69 "bi_db.m4"
install_pragma("dbprop", 3, 4, _pragma_dbprop);
#line 197 "bi_db.m4"
va_builtin_install_ex("dbmap", bi_dbmap, 0, dtype_number, 3, 1, 0|0, dtype_string, dtype_string, dtype_number);
#line 207 "bi_db.m4"
va_builtin_install_ex("dbget", bi_dbget, 0, dtype_string, 4, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number);
#line 218 "bi_db.m4"
va_builtin_install_ex("dbput", bi_dbput, 0, dtype_unspecified, 5, 2, 0|0, dtype_string, dtype_string, dtype_string, dtype_number, dtype_number);
#line 256 "bi_db.m4"
va_builtin_install_ex("dbdel", bi_dbdel, 0, dtype_unspecified, 4, 2, 0|0, dtype_string, dtype_string, dtype_number, dtype_number);
#line 322 "bi_db.m4"
DBTAB_id = builtin_priv_register(alloc_db_tab, destroy_db_tab,
#line 322
NULL);
#line 338 "bi_db.m4"
va_builtin_install_ex("dbfirst", bi_dbfirst, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 366 "bi_db.m4"
va_builtin_install_ex("dbnext", bi_dbnext, 0, dtype_number, 1, 0, 0|0, dtype_number);
#line 387 "bi_db.m4"
va_builtin_install_ex("dbkey", bi_dbkey, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 405 "bi_db.m4"
va_builtin_install_ex("dbvalue", bi_dbvalue, 0, dtype_string, 1, 0, 0|0, dtype_number);
#line 443 "bi_db.m4"
install_pragma("greylist", 2, 2, _pragma_greylist);
#line 489 "bi_db.m4"
	greylist_seconds_left_loc = builtin_variable_install("greylist_seconds_left", dtype_number, VAR_VOLATILE)->off;
#line 751 "bi_db.m4"
va_builtin_install_ex("greylist", bi_greylist, 0, dtype_number, 2, 0, 0|0, dtype_string, dtype_number);
#line 763 "bi_db.m4"
va_builtin_install_ex("is_greylisted", bi_is_greylisted, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 769 "bi_db.m4"
va_builtin_install_ex("db_name", bi_db_name, 0, dtype_string, 1, 0, 0|0, dtype_string);
#line 779 "bi_db.m4"
va_builtin_install_ex("db_get_active", bi_db_get_active, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 789 "bi_db.m4"
va_builtin_install_ex("db_set_active", bi_db_set_active, 0, dtype_unspecified, 2, 0, 0|0, dtype_string, dtype_number);
#line 799 "bi_db.m4"
va_builtin_install_ex("db_expire_interval", bi_db_expire_interval, 0, dtype_number, 1, 0, 0|0, dtype_string);

#line 809 "bi_db.m4"
		
#line 809
}

#line 610 "../mfd/snarf.m4"

#line 610

